import java.util.LinkedList;

public class Riddle {
	enum Type {ADD, SUB, MUL, DIV, EQ, VAL};
	class Node {
		boolean visited = false;
		Type type;
		int val;
		boolean result;
		LinkedList<Node> neighbours;
		
		public Node(int val) {
			this.val = val;
			type = Type.VAL;			
		}
		
		public Node(Type type) {
			this.type = type;			
		}
		
		@Override
		public String toString() {
			switch(type) {
			case ADD: return "+";
			case SUB: return "-";
			case MUL: return "x";
			case DIV: return ":";
			case EQ: return "=";
			default: return ""+val;
			}
		}
	}
	
	boolean isSolution(LinkedList<Node> p) {
		int tmpRes = p.get(0).val;
		for(int i = 1; i < p.size() - 1; i+= 2) {
			Node op = p.get(i); Node v = p.get(i+1);
			switch(op.type) {
			case ADD: tmpRes += v.val; break;
			case SUB: tmpRes -= v.val; break;
			case MUL: tmpRes *= v.val; break;
			case DIV:
				if(tmpRes % v.val != 0) return false;
				tmpRes /= v.val; break;
			case EQ:
				if(v.result && tmpRes == v.val) return true;
				return false;
			default: return false;
			}
		}
		
		return false;
	}
	
	void addList(LinkedList<Node> p, LinkedList<LinkedList<Node>> s) {
		s.addLast(new LinkedList<>(p));
	}
	
	void allPaths(Node n, LinkedList<Node> p, LinkedList<LinkedList<Node>> s) {
		if(n.visited) return;
		n.visited = true;
		
		if(n.result) {
			p.addLast(n);
			addList(p, s);
			p.removeLast();
			n.visited = false;
		} else {
			p.addLast(n);
			for(Node neighbour : n.neighbours)
				allPaths(neighbour, p, s);
			p.removeLast();
			n.visited = false;
		}
	}
	
	LinkedList<Node> solve(Node start) {
		LinkedList<LinkedList<Node>> s = new LinkedList<>();
		allPaths(start, new LinkedList<Node>(), s);
		
		LinkedList<Node> best = null;
		for(LinkedList<Node> path : s) {
			if(isSolution(path)) {
				if(best == null || path.size() < best.size())
					best = path;
			}
		}
		
		return best;
	}
	
	LinkedList<Node> buildRiddle(String[][] mat) {
		Node[][] nodes = new Node[mat.length][mat[0].length];
		LinkedList<Node> startEnd = new LinkedList<>();
		
		for(int i = 0; i < nodes.length; i++)
			for(int j = 0; j < nodes[0].length; j++)
				nodes[i][j] = decode(mat[i][j], startEnd);
		
		for(int i = 0; i < nodes.length; i++)
			for(int j = 0; j < nodes[0].length; j++) {
				LinkedList<Node> neighbours = new LinkedList<>();
				
				if(i > 0 && nodes[i - 1][j] != null)
					neighbours.add(nodes[i - 1][j]);
				if(i < nodes.length - 1 && nodes[i + 1][j] != null)
					neighbours.add(nodes[i + 1][j]);
				if(j > 0 && nodes[i][j - 1] != null)
					neighbours.add(nodes[i][j - 1]);
				if(j < nodes[0].length - 1 && nodes[i][j + 1] != null)
					neighbours.add(nodes[i][j + 1]);
				
				if(nodes[i][j] != null)
					nodes[i][j].neighbours = neighbours;
			}
		
		return startEnd;
	}
	
	private Node decode(String s, LinkedList<Node> startEnd) {		
		switch(s) {
		case "": return null;
		case "+": return new Node(Type.ADD);
		case "-": return new Node(Type.SUB);
		case "x": return new Node(Type.MUL);
		case ":": return new Node(Type.DIV);
		case "=": return new Node(Type.EQ);
		}
		
		int value = 0;
		if(s.startsWith("!")) {
			value = Integer.parseInt(s.substring(1));
			Node n = new Node(value);
			startEnd.add(n);
			return n;
		} else if(s.startsWith("?")) {
			value = Integer.parseInt(s.substring(1));
			Node n = new Node(value);
			n.result = true;
			startEnd.add(n);
			return n;
		} else {
			value = Integer.parseInt(s);
			return new Node(value);
		}
	}
	
	public static void main(String[] args) {
		String[][] riddle = new String[][] {
			{"", "4", "x", "3", "-"},
			{"1", "+", "!6", "+", "2"},
			{"=", "7", "+", "5", ":"},
			{"2", "-", "7", ":", "2"},
			{"-", "3", "=", "3", "+"},
			{"8", "=", "?5", "x", "2"},			
		};
		
		Riddle r = new Riddle();
		LinkedList<Node> startEnd = r.buildRiddle(riddle);
		
		LinkedList<Node> solution = r.solve(startEnd.get(0));
		
		for(Node n : solution)
			System.out.println(n);
	}
}
